//
//  StatusActionAccountListCollectionViewController.swift
//  Tusker
//
//  Created by Shadowfacts on 9/5/19.
//  Copyright © 2019 Shadowfacts. All rights reserved.
//

import UIKit
import Pachyderm

class StatusActionAccountListCollectionViewController: UIViewController, CollectionViewController {
    
    private let mastodonController: MastodonController
    
    /// If `true`, a warning will be shown below the account list describing that the total favs/reblogs may be innacurate.
    var showInacurateCountWarning = false
    
    var collectionView: UICollectionView! {
        view as? UICollectionView
    }
    private var dataSource: UICollectionViewDiffableDataSource<Section, Item>!
    
    /**
     Creates a new view controller showing the accounts that performed the given action on the given status.
     
     - Parameter mastodonController The `MastodonController` instance this view controller uses.
     */
    init(mastodonController: MastodonController) {
        self.mastodonController = mastodonController
        
        super.init(nibName: nil, bundle: nil)
        
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func loadView() {
        let layout = UICollectionViewCompositionalLayout { [unowned self] sectionIndex, environment in
            switch dataSource.sectionIdentifier(for: sectionIndex)! {
            case .status:
                var config = UICollectionLayoutListConfiguration(appearance: .grouped)
                config.footerMode = self.showInacurateCountWarning ? .supplementary : .none
                config.leadingSwipeActionsConfigurationProvider = { [unowned self] in
                    (collectionView.cellForItem(at: $0) as? TimelineStatusCollectionViewCell)?.leadingSwipeActions()
                }
                config.trailingSwipeActionsConfigurationProvider = { [unowned self] in
                    (collectionView.cellForItem(at: $0) as? TimelineStatusCollectionViewCell)?.trailingSwipeActions()
                }
                return NSCollectionLayoutSection.list(using: config, layoutEnvironment: environment)
            case .accounts:
                return NSCollectionLayoutSection.list(using: .init(appearance: .grouped), layoutEnvironment: environment)
            }
        }
        view = UICollectionView(frame: .zero, collectionViewLayout: layout)
        collectionView.delegate = self
        collectionView.dragDelegate = self
        collectionView.allowsFocus = true
        dataSource = createDataSource()
    }
    
    private func createDataSource() -> UICollectionViewDiffableDataSource<Section, Item> {
        let statusCell = UICollectionView.CellRegistration<TimelineStatusCollectionViewCell, (String, CollapseState)> { [unowned self] cell, indexPath, item in
            cell.delegate = self
            cell.updateUI(statusID: item.0, state: item.1, filterResult: .allow, precomputedContent: nil)
        }
        let accountCell = UICollectionView.CellRegistration<AccountCollectionViewCell, String> { [unowned self] cell, indexPath, item in
            cell.delegate = self
            cell.updateUI(accountID: item)
        }
        let dataSource = UICollectionViewDiffableDataSource<Section, Item>(collectionView: collectionView) { collectionView, indexPath, itemIdentifier in
            switch itemIdentifier {
            case .status(let id, let state):
                return collectionView.dequeueConfiguredReusableCell(using: statusCell, for: indexPath, item: (id, state))
            case .account(let id):
                return collectionView.dequeueConfiguredReusableCell(using: accountCell, for: indexPath, item: id)
            }
        }
        let sectionHeaderCell = UICollectionView.SupplementaryRegistration<UICollectionViewListCell>(elementKind: UICollectionView.elementKindSectionFooter) { (headerView, collectionView, indexPath) in
            var config = headerView.defaultContentConfiguration()
            config.text = NSLocalizedString("Favorite and reblog counts for posts originating from instances other than your own may not be accurate.", comment: "shown on lists of status total actions")
            headerView.contentConfiguration = config
        }
        dataSource.supplementaryViewProvider = { collectionView, elementKind, indexPath in
            return collectionView.dequeueConfiguredReusableSupplementary(using: sectionHeaderCell, for: indexPath)
        }
        return dataSource
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        clearSelectionOnAppear(animated: animated)
    }
    
    func addStatus(_ status: StatusMO, state: CollapseState) {
        loadViewIfNeeded()
        
        var snapshot = NSDiffableDataSourceSnapshot<Section, Item>()
        snapshot.appendSections([.status, .accounts])
        snapshot.appendItems([.status(status.id, state)], toSection: .status)
        dataSource.apply(snapshot, animatingDifferences: false)
    }
    
    func addAccounts(_ accountIDs: [String], animated: Bool) {
        var snapshot = dataSource.snapshot()
        snapshot.appendItems(accountIDs.map { .account($0) }, toSection: .accounts)
        dataSource.apply(snapshot, animatingDifferences: animated)
    }
    
}

extension StatusActionAccountListCollectionViewController {
    enum Section {
        case status
        case accounts
    }
    enum Item: Hashable {
        case status(String, CollapseState)
        case account(String)
        
        static func ==(lhs: Item, rhs: Item) -> Bool {
            switch (lhs, rhs) {
            case (.status(let a, _), .status(let b, _)):
                return a == b
            case (.account(let a), .account(let b)):
                return a == b
            default:
                return false
            }
        }
        
        func hash(into hasher: inout Hasher) {
            switch self {
            case .status(let id, _):
                hasher.combine(0)
                hasher.combine(id)
            case .account(let id):
                hasher.combine(1)
                hasher.combine(id)
            }
        }
    }
}

extension StatusActionAccountListCollectionViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        switch dataSource.itemIdentifier(for: indexPath) {
        case nil:
            return
        case .status(let id, let state):
            selected(status: id, state: state.copy())
        case .account(let id):
            selected(account: id)
        }
    }
    
    func collectionView(_ collectionView: UICollectionView, contextMenuConfigurationForItemAt indexPath: IndexPath, point: CGPoint) -> UIContextMenuConfiguration? {
        guard let item = dataSource.itemIdentifier(for: indexPath),
              let cell = collectionView.cellForItem(at: indexPath) else {
            return nil
        }
        switch item {
        case .status:
            return (cell as? TimelineStatusCollectionViewCell)?.contextMenuConfiguration()
        case .account(let id):
            return UIContextMenuConfiguration {
                ProfileViewController(accountID: id, mastodonController: self.mastodonController)
            } actionProvider: { _ in
                UIMenu(children: self.actionsForProfile(accountID: id, source: .view(cell)))
            }
        }
    }
    
    func collectionView(_ collectionView: UICollectionView, willPerformPreviewActionForMenuWith configuration: UIContextMenuConfiguration, animator: UIContextMenuInteractionCommitAnimating) {
        MenuPreviewHelper.willPerformPreviewAction(animator: animator, presenter: self)
    }
}

extension StatusActionAccountListCollectionViewController: UICollectionViewDragDelegate {
    func collectionView(_ collectionView: UICollectionView, itemsForBeginning session: UIDragSession, at indexPath: IndexPath) -> [UIDragItem] {
        guard let currentAccountID = mastodonController.accountInfo?.id,
              let item = dataSource.itemIdentifier(for: indexPath) else {
            return []
        }
        let provider: NSItemProvider
        switch item {
        case .status(let id, _):
            guard let status = mastodonController.persistentContainer.status(for: id) else {
                return []
            }
            provider = NSItemProvider(object: status.url! as NSURL)
            let activity = UserActivityManager.showConversationActivity(mainStatusID: id, accountID: currentAccountID)
            activity.displaysAuxiliaryScene = true
            provider.registerObject(activity, visibility: .all)
        case .account(let id):
            guard let account = mastodonController.persistentContainer.account(for: id) else {
                return []
            }
            provider = NSItemProvider(object: account.url as NSURL)
            let activity = UserActivityManager.showProfileActivity(id: account.id, accountID: currentAccountID)
            activity.displaysAuxiliaryScene = true
            provider.registerObject(activity, visibility: .all)
        }
        return [UIDragItem(itemProvider: provider)]
    }
}

extension StatusActionAccountListCollectionViewController: TuskerNavigationDelegate {
    var apiController: MastodonController! { mastodonController }
}

extension StatusActionAccountListCollectionViewController: MenuActionProvider {
}

extension StatusActionAccountListCollectionViewController: StatusCollectionViewCellDelegate {
    func statusCellNeedsReconfigure(_ cell: StatusCollectionViewCell, animated: Bool, completion: (() -> Void)?) {
        if let indexPath = collectionView.indexPath(for: cell) {
            var snapshot = dataSource.snapshot()
            snapshot.reconfigureItems([dataSource.itemIdentifier(for: indexPath)!])
            dataSource.apply(snapshot, animatingDifferences: animated, completion: completion)
        }
    }
    
    func statusCellShowFiltered(_ cell: StatusCollectionViewCell) {
        fatalError()
    }
}

extension StatusActionAccountListCollectionViewController: StatusBarTappableViewController {
    func handleStatusBarTapped(xPosition: CGFloat) -> StatusBarTapActionResult {
        collectionView.scrollToTop()
        return .stop
    }
}
